# -*- coding: utf-8 -*-

#  format of "/proc/net/tcpprobe"
#
#  11.30814 192.168.123.42:35120 192.168.123.99:80 1472 0x259a6f5 0x259a6f5 10 2147483647 19840 36604 150784
#  ^        ^                    ^                ^    ^         ^         ^  ^          ^     ^     ^
#  |        |                    |                |    |         |         |  |          |     |     +-------- [10] RWND
#  |        |                    |                |    |         |         |  |          |     +--------------  [9] sRTT (!) µs
#  |        |                    |                |    |         |         |  |          +--------------------  [8] Send window
#  |        |                    |                |    |         |         |  +-------------------------------  [7] Slow start threshold
#  |        |                    |                |    |         |         +----------------------------------  [6] Congestion window
#  |        |                    |                |    |         +--------------------------------------------  [5] Unacknowledged sequence #
#  |        |                    |                |    +------------------------------------------------------  [4] Next send sequence #
#  |        |                    |                +-----------------------------------------------------------  [3] Bytes in packet
#  |        |                    +----------------------------------------------------------------------------  [2] Receiver address:port
#  |        +-------------------------------------------------------------------------------------------------  [1] Sender address:port
#  +----------------------------------------------------------------------------------------------------------  [0] Time seconds

import sys
import os
import select
import subprocess
from ipaddress import ip_address

from ...utils.utilty import Utility
from .input_base import InputBase
from ...sample import Sample

PATH_TO_TCP_PROBE = "/proc/net/tcpprobe"
TCP_PROBE_FORMAT = ['time', 'src', 'dst', 'packetSize', 'seq', 'unackSeq', 'cwnd', 'sst', 'swnd', 'rtt', 'rwnd']
TCP_PROBE_POLL_TIMEOUT = 0.01

class TcpProbeInput(InputBase):
    def __init__(self, options, infoRegistry):
        self.options = options
        self.infoRegistry = infoRegistry
        self.__linesProccessed = 0
        self.__matchedLinesProccessed = 0
        self.closed = False

    def startupCheck(self):
        readable = os.access(PATH_TO_TCP_PROBE, os.R_OK)
        if(not readable):
            Utility.eprint(\
                "Error: TCP-Probe file \"" + PATH_TO_TCP_PROBE + "\" not readable. Exiting..."\
                "\n\n"\
                "Kernel module loaded? Permissions set? Try:\n"\
                "sudo modprobe tcp_probe full=1 port=0 && sudo chmod 444 /proc/net/tcpprobe"\
                "\n\n"\
                "OR use \"tcpinfo\" as input-backend instead"\
                )
            sys.exit(1)

        # count stdout of lsof and do some dirty casting...
        command = "lsof -t " + PATH_TO_TCP_PROBE + "|wc -l"
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=None, shell=True)
        numberOfProcsAsString = bytes(process.communicate()[0]).decode('utf-8')
        numberOfProcs = int(numberOfProcsAsString)
        if(numberOfProcs > 0):
            Utility.eprint("Error: Input file already in use by another process. Exiting...")
            sys.exit(1)

    def startUp(self):
        self.__tcpProbeFileHandler = open(PATH_TO_TCP_PROBE, "r")

    def tearDown(self):
        self.closed = True
        self.__tcpProbeFileHandler.close()

    def retrieveNewSamples(self):
        newSamples = []
        if(self.closed):
            return newSamples

        inputReady,outputReady,exceptReady = select.select([self.__tcpProbeFileHandler.fileno()], [], [], TCP_PROBE_POLL_TIMEOUT)
        if(inputReady):
            for s in inputReady:
                if(self.closed):
                    return newSamples
                elif(s == self.__tcpProbeFileHandler.fileno()):
                    try:
                        line = self.__tcpProbeFileHandler.readline()
                        proccessSample = self.processLogLine(line)
                        newSamples.append(proccessSample)

                    except IOError:
                        pass

                    else:
                        return newSamples
        return newSamples

    def processLogLine(self, line):
        self.__linesProccessed += 1
        self.__matchedLinesProccessed += 1
        log = False
        tmpData = line.strip().split(" ")
        data = dict(zip(TCP_PROBE_FORMAT, tmpData))

        srcIp, separator, srcPort = data['src'].rpartition(':')
        assert separator
        srcPort = int(srcPort)
        srcIp = ip_address(srcIp.strip("[]"))

        dstIp, separator, dstPort = data['dst'].rpartition(':')
        assert separator
        dstPort = int(dstPort)
        dstIp = ip_address(dstIp.strip("[]"))

        if(self.options.filterWhiteListing):
            log = False
            if((srcPort in self.options.filterSrcPorts) or (dstPort in self.options.filterDstPorts)):
                log = True

        else: # blacklisting
            log = True
            if(srcPort in self.options.filterSrcPorts or dstPort in self.options.filterDstPorts):
                log = False


        # line matches - continue processing
        if(log):
            self.__matchedLinesProccessed += 1

            # time since opening of tcp-probe file
            relativeTimestamp = float(data['time'])
            rtt = float(data['rtt']) / 1000 # now rtt is in ms
            seq = int(data['seq'], 0)

            # create ne sample object and return it
            sample = Sample(srcIp, srcPort, dstIp, dstPort)
            sample.relativeTimestamp = relativeTimestamp
            sample.cwnd = int(data['cwnd'])
            sample.rwnd = int(data['rwnd'])
            sample.sst = int(data['sst'])
            sample.seq = seq
            sample.rtt = rtt

            return sample
        else:
            return None



# -*- coding: utf-8 -*-

import sys
import threading
import time
# import netifaces
# from netifaces import AF_INET, AF_INET6
from collections import deque
from collections import OrderedDict

from .flow import Flow

THREAD_STOPFLAG_WAIT = 0.000001 # in seconds
THREAD_JOIN_TIMEOUT = 1 # in seconds
THREAD_TESTTIMEOUTS_WAIT = 0.5 # in seconds
THREAD_MISC_WAIT = 2 # in seconds

class TcpLog():
    def __init__(self, inputBackend, outputBackend, guiBackend, options, infoRegistry):
        self.inputBackend = inputBackend
        self.outputBackend = outputBackend
        self.guiBackend = guiBackend
        self.options = options
        self.infoRegistry = infoRegistry

    def run(self):
        """Initializes data structures and start worker-threads"""

        self.__stopped = threading.Event()

        self.startTimestamp = self.infoRegistry.load('startTime')
        self.timestampLastSample = 0
        self.numberOfLoggedSamples = 0

        # self.ownIpAdresses = self.determineOwnIpAdresses()
        # self.infoRegistry.save('ownIpAdresses', self.ownIpAdresses)
        self.flows = OrderedDict()

        self.__processInputThread = None
        self.__processOutputThread = None
        self.__processGuiThread = None
        self.__autoStopControlThread = None
        # self.__processMiscThread = None

        # init threads
        self.__processInputThread = threading.Thread(target=self.processInput)
        self.__processInputThread.daemon = True
        self.__processOutputThread = threading.Thread(target=self.processOutput)
        self.__processOutputThread.daemon = True

        # process misc. stuff --> utility thread
        # self.__processMiscThread = threading.Thread(target=self.processMisc)
        # self.__processMiscThread.daemon = True

        if(self.options.quiet != True and self.options.outputBackend != "stdout"):
            self.__processGuiThread = threading.Thread(target=self.processGui)
            self.__processGuiThread.daemon = True
            self.__processGuiThread.start()

        if(self.options.inActivityTimeout > 0 or self.options.hardTimeout > 0):
            self.__autoStopControlThread = threading.Thread(target=self.testAutoTimeouts)
            self.__autoStopControlThread.daemon = True
            self.__autoStopControlThread.start()

        self.__processInputThread.start()
        self.__processOutputThread.start()
        # self.__processMiscThread.start()

        while(not self.__stopped.wait(THREAD_STOPFLAG_WAIT)):
            self.__stopped.wait(0.5)

        if(self.options.debug):
            print("End of main thread reached.")

    def processInput(self):
        while(not self.__stopped.wait(THREAD_STOPFLAG_WAIT)):
            newSamples = self.inputBackend.retrieveNewSamples()
            for sampleCandidate in newSamples:
                if(sampleCandidate is not None):
                    flowIdentifier = sampleCandidate.getFlowIdentifier()
                    if flowIdentifier in self.flows:
                        # known flow
                        thisFlow = self.flows[flowIdentifier]
                        if (sampleCandidate.absoluteTimestamp - thisFlow.currentSample.absoluteTimestamp) > self.options.logResolution:
                            # log sample
                            self.flows[flowIdentifier].updateFlow(sampleCandidate)
                            self.updateGlobalStats(sampleCandidate)
                        else:
                            # got already a sample within logRes-timespan: skip sample
                            pass
                    else:
                        # flow not yet tracked: add to list of flows
                        self.flows[flowIdentifier] = Flow(flowIdentifier, sampleCandidate.srcIp, sampleCandidate.srcPort, sampleCandidate.dstIp, sampleCandidate.dstPort, sampleCandidate)
                        self.updateGlobalStats(sampleCandidate)
                        pass

            # wakeup several times within a logRes-timespan --> try not to miss new samples and to "flush" tcpprobe
            self.__stopped.wait(self.options.samplePollWait)
        pass

    def updateGlobalStats(self, sample):
        if(sample.absoluteTimestamp > self.timestampLastSample):
            self.timestampLastSample = sample.absoluteTimestamp
        self.numberOfLoggedSamples += 1

    def processOutput(self):
        if(self.outputBackend is None):
            return
        sampleOutputQueue = deque()
        self.outputBackend.setSampleQueue(sampleOutputQueue)
        self.outputBackend.setValuesToLog(self.options.valuesToLog)

        self.outputBackend.preLogging()

        while(not self.__stopped.wait(THREAD_STOPFLAG_WAIT)):
            for flow in self.flows:
                thisFlow = self.flows[flow]
                if(thisFlow.isActive()):
                    if(self.options.fillSampleGaps or not thisFlow.getLogStatus()):
                        sampleOutputQueue.append(thisFlow.currentSample)
                        thisFlow.setLogStatus(True)

            self.outputBackend.logSample()

            if(self.options.fillSampleGaps):
                self.__stopped.wait(self.options.logResolution)

        self.outputBackend.postLogging()

    def processGui(self):
        if(self.options.guiBackend == "curses"):
            self.guiBackend.setFlows(self.flows)

        self.__stopped.wait(1) # loading...

        while(not self.__stopped.wait(THREAD_STOPFLAG_WAIT)):
            self.guiBackend.refresh()
            if(self.guiBackend.isUnloadingRequested()):
                self.initTearDown()
                self.guiBackend.tearDown()
                return

        if(self.options.guiBackend is not None):
            self.guiBackend.tearDown()

    def testAutoTimeouts(self):
        # TODO: print info about timeout-event --> IF not stdout/quiet
        while(not self.__stopped.wait(THREAD_STOPFLAG_WAIT)):
            currentTimestamp = time.time()
            if(self.options.hardTimeout > 0):
                elapsedTime = currentTimestamp - self.startTimestamp
                self.infoRegistry.save("hardTimeout_elapsedTime", elapsedTime)
                if(elapsedTime >= self.options.hardTimeout):
                    # hard-timeout reached --> exiting
                    self.initTearDown()
                    print("Hard-Timeout reached: Auto exiting.")
                    return

            elif(self.options.inActivityTimeout > 0):
                if(self.timestampLastSample > 0):
                    timeSinceLastFlowActivity = currentTimestamp - self.timestampLastSample
                else:
                    timeSinceLastFlowActivity = -1
                self.infoRegistry.save("softTimeout_timeSinceLastFlowActivity", timeSinceLastFlowActivity)
                if(timeSinceLastFlowActivity > self.options.inActivityTimeout):
                    if(self.timestampLastSample > 0):
                        # all flows inactive for at leaset "inActivityTimeout" --> exiting
                        self.initTearDown()
                        print("InActivity-Timeout reached: Auto exiting.")
                        return

            else:
                self.__stopped.wait(THREAD_TESTTIMEOUTS_WAIT)

    # def processMisc(self):
    #     while(not self.__stopped.wait(THREAD_STOPFLAG_WAIT)):
    #         # determine FQDN outside of main-threads b/c lookup can take a while
    #         if(self.options.translateIp2Hostname):
    #             for flow in self.flows:
    #                 thisFlow = self.flows[flow]
    #                 if(thisFlow.srcHostname is None):
    #                     thisFlow.srcHostname = socket.getfqdn(thisFlow.srcIp)
    #                 if(thisFlow.dstHostname is None):
    #                     thisFlow.dstHostname = socket.getfqdn(thisFlow.dstIp)
    #         self.__stopped.wait(THREAD_MISC_WAIT)

    # def determineOwnIpAdresses(self):
    #     ownIpAdresses = []
    #     for interface in netifaces.interfaces():
    #         try:
    #             ip4addr = netifaces.ifaddresses(interface)[AF_INET][0]['addr']
    #         except KeyError:
    #             pass
    #         else:
    #             if(len(ip4addr) > 0):
    #                 ownIpAdresses.append(ip4addr)
    #
    #         try:
    #             tmpIp6addr = netifaces.ifaddresses(interface)[AF_INET6][0]['addr']
    #             # returns for example:  e20::123:d9bb:feca:6353%wlp3s0
    #         except KeyError:
    #             pass
    #         else:
    #             if(len(tmpIp6addr) > 0):
    #                 if("%" in tmpIp6addr):
    #                     ip6addr, seperator, interface = tmpIp6addr.rpartition("%")
    #                     if(len(ip6addr) > 0):
    #                         ownIpAdresses.append(ip6addr)
    #                 else:
    #                     ownIpAdresses.append(tmpIp6addr)
    #     return ownIpAdresses

    def handleSignals(self, signal, frame):
        """Callback handler for signals"""
        if(not self.options.quiet and not self.options.outputBackend == 'stdout'):
            print("Exiting...")
        self.initTearDown()
        self.tearDown()
        raise SystemExit
        sys.exit(0)

    def initTearDown(self):
        self.__stopped.set()

    def tearDown(self):
        """Performs the cleanup at programm termination."""

        if(self.inputBackend is not None):
            self.inputBackend.tearDown()
        if(self.outputBackend is not None):
            self.outputBackend.tearDown()

        self.__processInputThread.join(THREAD_JOIN_TIMEOUT)
        self.__processOutputThread.join(THREAD_JOIN_TIMEOUT)
        if(self.__processGuiThread is not None):
            self.guiBackend.tearDown()
            self.__processGuiThread.join(THREAD_JOIN_TIMEOUT)
        if(self.__autoStopControlThread is not None):
            self.__autoStopControlThread.join(THREAD_JOIN_TIMEOUT)

        raise SystemExit
        sys.exit(0)



# -*- coding: utf-8 -*-

class InfoRegistry(object):
    """Holds global information to be shared between threads"""

    def __init__(self):
        """Does nothing"""
        self.__registry = {}

    def save(self, key, value):
        self.__registry[key] = value

    def load(self, key):
        if key in self.__registry:
            return self.__registry[key]
        else:
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Convenience wrapper for running TCPlog directly from source tree."""

import sys
sys.settrace

from tcplog.main import main

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
from ..backend_base import BackendBase

class GuiBase(BackendBase, metaclass=ABCMeta):
    @abstractmethod
    def repaint(self):
        """
        Repaints gui with all elements.
        """
        pass

    @abstractmethod
    def refresh(self):
        """
        Processes user input and calls rapaint if need be.
        """
        pass

    @abstractmethod
    def initTearDown(self):
        """
        Initializes a teardown and alerts user if need be.
        """
        pass



# -*- coding: utf-8 -*-

# TODO: cleaner?
STATS_TO_PRINT = ['#', 'activity', 'elapsed', 'srcIp', 'srcPort', 'dstIp', 'dstPort', 'cwnd', 'sst', 'rtt', 'minRtt', 'maxRtt', 'avgRtt', 'meanRtt', 'smoothedThroughput', 'assumedLosses']
STATS_TO_PRINT_TITLES = ['#', 'activity', 'elapsed', 'locIp', 'loc. port', 'remote IP', 'rem. port', 'cwnd', 'sst', 'rtt', 'minRtt', 'maxRtt', 'avgRtt', 'meanRtt', 'throughput', 'losses']
STATS_TO_PRINT_UNITS = [' ', 's', 's', 'ip', 'port', 'ip', 'port', 'pkt', 'pkt', 'ms', 'ms', 'ms', 'ms', 'ms', 'mbit/s', '#']
# STATS_TO_PRINT_UNITS_TYPE = ['time', 'time', 'str', 'str', 'str', 'str', 'count', 'count', 'time', 'time', 'time', 'time', 'time', 'transferSpeed', 'count']
# STATS_TO_PRINT_TITLES = [...... avgRtt, meanRtt, bw, aLosses]

import curses
import math
import time

from ...utils.utilty import Utility
from .gui_base import GuiBase

NUMBER_OF_DECIMALS = 2
MAX_INT_32 = 2147483647 # initial sst value...
MIN_TERM_WIDTH = 68
MIN_TERM_HEIGHT = 10
VERTICAL_SEPARATOR = " "
HORIZONTAL_SEPARATOR = "_"
COLORS_ENABLED = True
KEYCODE_ESC = 27
REFRESH_RATE_STEP = 0.5
MIN_REFRESH_RATE = 0.5
SCROLL_STEP_X = 5
SCROLL_STEP_Y = 1
SCROLL_STEP_Y_LARGE = 10 # PageUp/Down
KEYSTROKE_COOLDOWN = 0.100 # ms until next keystroke is processed
FLOW_VALUE_CELL_WIDTH = 11
INNER_PAD_SUB = 4 # 2 rows for header and 2 rows for footer

class CursesGui(GuiBase):
    def __init__(self, options, infoRegistry):
        self.options = options
        self.infoRegistry = infoRegistry
        self.window = None
        self.unloading = False
        self.outerPadHeight = 0
        self.outerPadWidth = 1000
        self.innerPadHeight = 0
        self.innerPadWidth = 1000
        self.timestampOfLastRedraw = 0
        self.timestrampOfLastKeystroke = 0
        self.innerPadPosCoordinates = 4, 0  # absolute position of pad
        self.innerPadStartCoordinates = 0, 0 # relative position of scroll-box
        self.outerPadPosCoordinates = 2, 0  # absolute position of pad
        self.outerPadStartCoordinates = 0, 0 # relative position of scroll-box

    def setFlows(self, flows):
        self.flows = flows

    def startupCheck(self):
        pass

    def startUp(self):
        self.window = curses.initscr()
        self.window.clear()
        self.window.nodelay(True)
        self.window.keypad(True)
        curses.savetty()
        curses.noecho()
        curses.cbreak()
        curses.curs_set(False)
        curses.start_color()
        curses.use_default_colors()

        if(COLORS_ENABLED):
            curses.init_pair(1, curses.COLOR_MAGENTA, -1)
            curses.init_pair(2, curses.COLOR_BLUE, -1)
            curses.init_pair(3, curses.COLOR_GREEN, -1)
            curses.init_pair(4, curses.COLOR_YELLOW, -1)
            curses.init_pair(5, curses.COLOR_CYAN, -1)
            curses.init_pair(6, curses.COLOR_WHITE, -1)
            curses.init_pair(7, curses.COLOR_RED, -1)
            curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_YELLOW)
        else:
            curses.init_pair(1, -1, -1)
            curses.init_pair(2, -1, -1)
            curses.init_pair(3, -1, -1)
            curses.init_pair(4, -1, -1)
            curses.init_pair(5, -1, -1)
            curses.init_pair(6, -1, -1)
            curses.init_pair(7, -1, -1)
            curses.init_pair(8, -1, -1)

        self.clearTerminal = True

        try:
            self.window.addstr(3, 3, "loading...", curses.A_BOLD)
            self.window.refresh()
            self.y, self.x = self.window.getmaxyx()
        except:
            print("Display Error! (Check terminal-size)")

    def isUnloadingRequested(self):
        return self.unloading

    def initTearDown(self):
        self.window.clear()
        self.window.addstr(3, 3, "unloading...", curses.A_BOLD)
        self.window.refresh()
        time.sleep(1)
        self.tearDown()
        self.unloading = True

    def tearDown(self):
        curses.nocbreak()
        curses.echo()
        curses.curs_set(True)
        self.window.keypad(False)
        curses.endwin()

    def processInput(self):
        self.processKeyInput()

    def processKeyInput(self):
        pressedkey = self.window.getch()
        if(pressedkey > 0):
            if(time.time() - self.timestrampOfLastKeystroke <= KEYSTROKE_COOLDOWN):
                curses.flushinp()
                return
            else:
                self.timestrampOfLastKeystroke = time.time()

                if(pressedkey == ord('q') or pressedkey == KEYCODE_ESC): # q or ESC
                    self.initTearDown()
                    return
                elif(pressedkey == curses.KEY_LEFT):
                    y,x = self.outerPadStartCoordinates
                    x -= SCROLL_STEP_X
                    x = max(0, x)
                    self.outerPadStartCoordinates = y, x

                    y,x = self.innerPadStartCoordinates
                    x -= SCROLL_STEP_X
                    x = max(0, x)
                    self.innerPadStartCoordinates = y, x
                    try:
                        self.outerPad.refresh(*self.outerPadStartCoordinates, *self.outerPadPosCoordinates, self.y-3, self.x-1)
                        self.innerPad.refresh(*self.innerPadStartCoordinates, *self.innerPadPosCoordinates, self.y-3, self.x-1)
                    except Exception as e:
                        pass
                elif(pressedkey == curses.KEY_RIGHT):
                    y,x = self.outerPadStartCoordinates
                    x += SCROLL_STEP_X
                    x = min(self.x, x)
                    self.outerPadStartCoordinates = y, x

                    y,x = self.innerPadStartCoordinates
                    x += SCROLL_STEP_X
                    x = min(self.x, x)
                    self.innerPadStartCoordinates = y, x
                    try:
                        self.outerPad.refresh(*self.outerPadStartCoordinates, *self.outerPadPosCoordinates, self.y-3, self.x-1)
                        self.innerPad.refresh(*self.innerPadStartCoordinates, *self.innerPadPosCoordinates, self.y-3, self.x-1)
                    except Exception as e:
                        pass
                elif(pressedkey == curses.KEY_UP or pressedkey == curses.KEY_PPAGE):
                    y,x = self.innerPadStartCoordinates
                    if(pressedkey == curses.KEY_UP):
                        y -= SCROLL_STEP_Y
                    elif(pressedkey == curses.KEY_PPAGE):
                        y -= SCROLL_STEP_Y_LARGE
                    y = max(0, y)
                    self.innerPadStartCoordinates = y, x
                    try:
                        self.innerPad.refresh(*self.innerPadStartCoordinates, *self.innerPadPosCoordinates, self.y-4, self.x-1)
                    except Exception as e:
                        pass
                elif(pressedkey == curses.KEY_DOWN or pressedkey == curses.KEY_NPAGE):
                    if((len(self.flows) + 2) > (self.y - INNER_PAD_SUB )):
                        y,x = self.innerPadStartCoordinates
                        if(pressedkey == curses.KEY_DOWN):
                            y += SCROLL_STEP_Y
                        elif(pressedkey == curses.KEY_NPAGE):
                            y += SCROLL_STEP_Y_LARGE
                        y = min(self.innerPadHeight - INNER_PAD_SUB, y)
                        try:
                            self.innerPadStartCoordinates = y, x
                            self.innerPad.refresh(*self.innerPadStartCoordinates, *self.innerPadPosCoordinates, self.y-4, self.x-1)
                        except Exception as e:
                            pass
                elif(pressedkey == ord('+')):
                    self.options.refreshRateLiveStats += REFRESH_RATE_STEP
                elif(pressedkey == ord('-')):
                    tmpValue = self.options.refreshRateLiveStats
                    tmpValue -= REFRESH_RATE_STEP
                    self.options.refreshRateLiveStats = max(MIN_REFRESH_RATE, tmpValue)

                self.repaint()



    def refresh(self):
        self.processInput()
        if(time.time() - self.timestampOfLastRedraw >= self.options.refreshRateLiveStats):
            self.timestampOfLastRedraw = time.time()
            self.repaint()
        else:
            time.sleep(0.1)

    def repaint(self):
        resize = curses.is_term_resized(self.y, self.x)
        if(resize is True):
            if(self.options.debug):
                self.window.addstr(3, 3, "Info: Terminal was resized.")
            self.y, self.x = self.window.getmaxyx()
            self.window.clear()
            curses.resizeterm(self.y, self.x)
            self.window.refresh()
            return

        if(self.x < MIN_TERM_WIDTH):
            self.window.clear()
            self.window.addstr(3, 3, "Error: Terminal too narrow.")
            self.window.refresh()
            self.clearTerminal = True
            return

        if(self.y < MIN_TERM_HEIGHT):
            self.window.clear()
            self.window.addstr(3, 3, "Error: Terminal too small.")
            self.window.refresh()
            self.clearTerminal = True
            return


        if(self.clearTerminal):
            self.window.clear()
            self.window.refresh()
            self.clearTerminal = False

        INFO_GRID_CELLS = 4
        INFO_GRID_CELL_WIDTH = math.floor((self.x-2) / INFO_GRID_CELLS)
        INFO_GRID_CELL_STARTS = {}
        for x in range(0, (INFO_GRID_CELLS)):
            INFO_GRID_CELL_STARTS[x] = (x * INFO_GRID_CELL_WIDTH) + 1

        try:

            # header area
            timeNow = time.strftime("%H:%M:%S")
            headerPos = 0 # + 1 @ border
            headerTexts = {}
            headerTexts[0] = 'TCPlog: ' + self.infoRegistry.load("version")
            headerTexts[1] = 'Time: {}'.format(timeNow)
            headerTexts[2] = 'Input: ' + self.options.inputBackend
            headerTexts[3] = 'Output: ' + self.options.outputBackend
            self.window.addnstr(headerPos, INFO_GRID_CELL_STARTS[0], headerTexts[0].ljust(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, curses.A_BOLD)
            self.window.addnstr(headerPos, INFO_GRID_CELL_STARTS[1], headerTexts[1].center(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, curses.A_BOLD)
            self.window.addnstr(headerPos, INFO_GRID_CELL_STARTS[2], headerTexts[2].center(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, curses.A_BOLD)
            self.window.addnstr(headerPos, INFO_GRID_CELL_STARTS[3], headerTexts[3].rjust(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, curses.A_BOLD)

            # footer area
            footerPos = self.y - 1 # -2 @ border

            # TODO: count active flows, count long-living, count all
            numberOfFlows = len(self.flows)
            numberOfActiveFlows = 0
            for flow in self.flows:
                if(self.flows[flow].isActive()):
                    numberOfActiveFlows += 1

            timeoutCellAttr = curses.A_BOLD
            hardTimeout = "N/A"
            if(self.options.hardTimeout > 0):
                elapsedTime = self.infoRegistry.load("hardTimeout_elapsedTime")
                hardTimeout = str(round(elapsedTime, 1))
                hardTimeout += "/"
                hardTimeout += str(self.options.hardTimeout) + "s"

                if(self.options.hardTimeout - elapsedTime <= 10):
                    timeoutCellAttr |= curses.A_BLINK

            softTimeout = "N/A"
            if(self.options.inActivityTimeout > 0):
                timeSinceLastFlowActivity = self.infoRegistry.load("softTimeout_timeSinceLastFlowActivity")
                softTimeout = str(round(timeSinceLastFlowActivity, 1))
                softTimeout += "/"
                softTimeout += str(self.options.inActivityTimeout) + "s"

                if(self.options.inActivityTimeout * 0.6 <= timeSinceLastFlowActivity):
                    timeoutCellAttr |= curses.A_BLINK

            footerTexts = {}
            footerTexts[0] = 'Interval: {}s'.format(round(self.options.refreshRateLiveStats, 1))
            footerTexts[1] = 'Log-Res: {}s'.format(self.options.logResolution)
            footerTexts[2] = 'Flows: ' + str(numberOfActiveFlows) + " (" + str(numberOfFlows) + ")"
            footerTexts[3] = 'ST: ' + str(softTimeout) + ' | HT: ' + str(hardTimeout)
            self.window.addnstr(footerPos, INFO_GRID_CELL_STARTS[0], footerTexts[0].ljust(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, curses.A_BOLD)
            self.window.addnstr(footerPos, INFO_GRID_CELL_STARTS[1], footerTexts[1].center(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, curses.A_BOLD)
            self.window.addnstr(footerPos, INFO_GRID_CELL_STARTS[2], footerTexts[2].center(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, curses.A_BOLD)
            self.window.addnstr(footerPos, INFO_GRID_CELL_STARTS[3], footerTexts[3].rjust(INFO_GRID_CELL_WIDTH), INFO_GRID_CELL_WIDTH, timeoutCellAttr)

            cellWidth = FLOW_VALUE_CELL_WIDTH
            normalCellWidth = cellWidth

            # TODO:
            # * print config before flow-info
            # * filter options
            # * inputFile, outputFile, logSocketPort
            # * stats --> samples processed, samples logged...
            # * numberOfSocket-clients
            # print value names

            # y-fixed content: header of flow-table --> only scrollable in x-direction
            self.outerPadHeight = 2 # header
            self.outerPad = curses.newpad(self.outerPadHeight, self.outerPadWidth)
            self.outerPad.scrollok(1)
            self.outerPad.idlok(1)
            # self.outerPad.bkgd("+")

            # print units
            startYindex = 1
            startXindex = 0
            i = 0
            for val in STATS_TO_PRINT_UNITS:
                cellWidth = normalCellWidth
                if(i is 0):
                    cellWidth = 5
                elif(i is 1):
                    cellWidth = 7
                elif(i is 2):
                    cellWidth = 5
                elif(i is 3):
                    i += 1
                    continue
                elif(i is 5):
                    cellWidth = 16
                tmpString = str(val)
                tmpString = tmpString[:(cellWidth-1)]
                self.outerPad.addnstr(startYindex, startXindex, str(tmpString).ljust(cellWidth), cellWidth)
                startXindex += cellWidth
                i += 1

            # print cell titles
            startYindex = 0
            startXindex = 0
            i = 0
            for val in STATS_TO_PRINT_TITLES:
                cellWidth = normalCellWidth
                if(i is 0):
                    cellWidth = 5
                elif(i is 1):
                    cellWidth = 7
                elif(i is 2):
                    cellWidth = 5
                elif(i is 3):
                    i += 1
                    continue
                elif(i is 5):
                    cellWidth = 16

                tmpString = str(val)
                tmpString = tmpString[:(cellWidth-1)]
                self.outerPad.addnstr(startYindex, startXindex, str(tmpString).ljust(cellWidth), cellWidth, curses.A_BOLD | curses.color_pair(8))
                startXindex += cellWidth
                i += 1



            self.innerPadHeight = len(self.flows)
            self.innerPadHeight = max(self.innerPadHeight, (self.y - INNER_PAD_SUB))
            self.innerPad = curses.newpad(self.innerPadHeight, self.innerPadWidth) # contains flows
            self.innerPad.scrollok(1)
            self.innerPad.idlok(1)
            # self.innerPad.bkgd("*")

            startYindex = 0
            startXindex = 0
            i = 0

            for flowKey, flowValue in self.flows.items():
                if(flowKey not in self.flows):
                    break;

                if(not self.options.showEphemeralFlows and self.flows[flowKey].isEphemeral()):
                    continue

                cellStyle = curses.color_pair(3)
                if(not self.flows[flowKey].isActive()):
                    cellStyle = curses.color_pair(4)

                for val in STATS_TO_PRINT:
                    cellWidth = normalCellWidth

                    if(val is '#'):
                        valueAsString = str(i)
                        cellWidth = 5

                    elif(val is 'activity'):
                        if(self.flows[flowKey].isActive()):
                            valueAsString = str(round(self.flows[flowKey].timeSinceLastActivity(), 2))
                        else:
                            valueAsString = "inact."
                        cellWidth = 7

                    elif(val is 'elapsed'):
                        valueAsString = str(int(round(self.flows[flowKey].totalActiveTime(), 0)))
                        cellWidth = 5

                    elif(val is 'srcIp'):
                        continue
                        if(self.flows[flowKey].srcHostname is None):
                            valueAsString = str(self.flows[flowKey].srcIp)
                        else:
                            valueAsString = str(self.flows[flowKey].srcHostname)

                    elif(val is 'dstIp'):
                        cellWidth = 16
                        if(self.flows[flowKey].dstHostname is None):
                            valueAsString = str(self.flows[flowKey].dstIp)
                        else:
                            valueAsString = str(self.flows[flowKey].dstHostname)


                    else:
                        valueAsString = "#"
                        sample = None
                        result = None
                        sample = self.flows[flowKey].currentSample
                        if(sample is not None):
                            result = getattr(sample, val, None)
                            if result is not None:
                                # maybe float?
                                if("." in str(result)):
                                    try:
                                        floatResult = float(result)
                                    except ValueError:
                                        # nope no float...
                                        valueAsString = str(result)
                                    else:
                                        valueAsString = str(Utility.truncate(floatResult, NUMBER_OF_DECIMALS))

                                # maybe int?
                                else:
                                    try:
                                        intResult = int(result)
                                    except ValueError:
                                        # nope no int...
                                        valueAsString = str(result)
                                    else:
                                        if(intResult == MAX_INT_32):
                                            valueAsString = "+INF"
                                        else:
                                            valueAsString = str(intResult)


                    valueAsString = valueAsString.ljust(cellWidth-1)
                    if(len(valueAsString) > cellWidth-1):
                        valueAsString = valueAsString[:(cellWidth-3)] + ".."

                    self.innerPad.addnstr(startYindex, startXindex, valueAsString, cellWidth, cellStyle)
                    startXindex += cellWidth
                startXindex = 0
                startYindex += 1
                i += 1

            self.outerPad.refresh(*self.outerPadStartCoordinates, *self.outerPadPosCoordinates, self.y-3, self.x-1)
            self.innerPad.refresh(*self.innerPadStartCoordinates, *self.innerPadPosCoordinates, self.y-3, self.x-1)
            self.window.refresh()

        except Exception as e:
            print(str(e))
            self.clearTerminal = True
            print("Error drawing live-stats. Terminal small?")

        return



# -*- coding: utf-8 -*-

from __future__ import print_function
import sys
import math

class Utility:
    @staticmethod
    def eprint(*args, **kwargs):
        print(*args, file=sys.stderr, **kwargs)

    @staticmethod
    def truncate(number, digits) -> float:
        stepper = pow(10.0, digits)
        return math.trunc(stepper * number) / stepper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os
import sys
import signal
import time

from .utils.utilty import Utility
from .tcplog import TcpLog
from .info_registry import InfoRegistry

# constants
TCP_LOG_VERSION = "0.2.7"
TCP_LOG_FORMAT_VERSION = "2"

# default values
DEFAULT_INPUT_BACKEND = "tcpinfo"
DEFAULT_OUTPUT_BACKEND = "none"
DEFAULT_REFRESH_RATE_LIVE_STATS = 1 # in seconds
DEFAULT_LOG_RESOLUTION = 0.1 # in seconds
DEFAULT_SOCKET_PORT = 11337
DEFAULT_LOGFILE_PATH = "/tmp/tcplog.log"
DEFAULT_VALUES_TO_LOG = ['absoluteTimestamp', 'relativeTimestamp', 'srcIp', 'srcPort', 'dstIp', 'dstPort', 'cwnd', 'sst', 'rtt', 'minRtt', 'maxRtt', 'avgRtt', 'meanRtt', 'throughput', 'smoothedThroughput', 'assumedLosses']


# main prog
def main():
    inputBackend = None
    outputBackend = None
    guiBackend = None

    infoRegistry = InfoRegistry()
    options = parse_options()
    startupSelfCheck(options)

    infoRegistry.save("startTime", time.time())
    infoRegistry.save("version", TCP_LOG_VERSION)
    infoRegistry.save("logFormatVersion", TCP_LOG_FORMAT_VERSION)

    if(options.showVersion):
        print("TCPlog: " + TCP_LOG_VERSION)
        print("Log format: " + TCP_LOG_FORMAT_VERSION)
        sys.exit(0)

    if(options.showGuiHelp):
        printGuiHelp()
        sys.exit(0)

    # init input backend
    if(options.inputBackend == "tcpinfo"):
        from .backends.input.tcpinfo import TcpInfoInput
        inputBackend = TcpInfoInput(options, infoRegistry)
        options.samplePollWait = options.logResolution / 10 # read often
    elif(options.inputBackend == "tcpprobe"):
        from .backends.input.tcpprobe import TcpProbeInput
        inputBackend = TcpProbeInput(options, infoRegistry)
        options.samplePollWait = options.logResolution / 1000 # try to read fast enough to flush tcpprobe
    # elif(options.inputBackend == "file"):
    #     from .backends.input.file import FileInput
    #     inputBackend = FileInput(options, info)
    #     options.samplePollWait = 0 # read file as fast as possible

    # init output backend
    if(options.outputBackend == "socket"):
        from .backends.output.socket import SocketOutput
        outputBackend = SocketOutput(options, infoRegistry)
    elif(options.outputBackend == "file"):
        from .backends.output.file import FileOutput
        outputBackend = FileOutput(options, infoRegistry)
    elif(options.outputBackend == "stdout"):
        from .backends.output.stdout import StdoutOutput
        outputBackend = StdoutOutput(options, infoRegistry)
    elif(options.outputBackend == "none"):
        pass

    # init gui backend
    if(options.guiBackend == "curses"):
        from .backends.gui.curses import CursesGui
        guiBackend = CursesGui(options, infoRegistry)
    # elif(options.guiBackend == "progress"):
    #     from .backends.gui.progress import ProgressGui
    #     guiBackend = ProgressGui(options, info)
    elif(options.guiBackend == None):
        guiBackend = None

    if(inputBackend is None):
        Utility.eprint("No valid input backend selected. Exiting...")
        sys.exit(1)

    if(inputBackend is not None):
        inputBackend.startupCheck()
        inputBackend.startUp()

    if(outputBackend is not None):
        outputBackend.startupCheck()
        outputBackend.startUp()

    if(guiBackend is not None):
        guiBackend.startupCheck()
        guiBackend.startUp()

    try:
        tcpLog = TcpLog(inputBackend, outputBackend, guiBackend, options, infoRegistry)
        signal.signal(signal.SIGINT, tcpLog.handleSignals)
        signal.signal(signal.SIGTERM, tcpLog.handleSignals)
        tcpLog.run()
        # ^-- starts the main programme
    except Exception as e:
        Utility.eprint(str(e))
        raise e
        sys.exit(1)
    finally:
        if(not options.quiet and not options.outputBackend == "stdout"):
            print("Goodbye cruel world!")
        sys.exit(0)


def parse_options():
    parser = argparse.ArgumentParser()

    # TODO:
    # * more filter options
    # * whitelisting
    # * blacklisting

    # parser.add_argument(
    #         "-a",
    #         "--notall", help="Filter flows by port (false)",
    #         action="store_false",
    #         dest="noFilter",
    #         default=True)

    # TODO: read from config file
    # parser.add_argument(
    #         "-c",
    #         "--config",
    #         help="Load configuration from file (default: None)",
    #         dest="configFile",
    #         default=None)

    parser.add_argument(
            "-r",
            "--log-resolution",
            help="Log resolution (in seconds, default: " + str(DEFAULT_LOG_RESOLUTION) + ")",
            dest="logResolution",
            type=float,
            default=DEFAULT_LOG_RESOLUTION)

    parser.add_argument(
            "-i",
            "--input-backend",
            help="Available input backends: \"tcpprobe\" and \"tcpinfo\" (default: " + DEFAULT_INPUT_BACKEND + ")",
            dest="inputBackend",
            default=DEFAULT_INPUT_BACKEND)

    parser.add_argument(
            "-o",
            "--output-backend",
            help="Available output backends: \"stdout\", \"file\", \"socket\" and \"none\" (default: " + DEFAULT_OUTPUT_BACKEND + ") ",
            dest="outputBackend",
            default=DEFAULT_OUTPUT_BACKEND)

    parser.add_argument(
            "-w",
            "--writeto",
            help="Path where the log file is stored - only usable with \"file\" output backend (default: " + DEFAULT_LOGFILE_PATH + ")",
            dest="logFilePath",
            default=DEFAULT_LOGFILE_PATH)

    parser.add_argument(
            "--log-port",
            help="Port number for socket-logging - only usable with \"socket\" output backend (default: " + str(DEFAULT_SOCKET_PORT) + ")",
            dest="logPort",
            type=int,
            default=DEFAULT_SOCKET_PORT)

    parser.add_argument(
            "--inactivity-timeout",
            help="SoftTimeout: Automatically stops logging if no flow activity is detected for n seconds (in seconds, default: 0 = disabled)",
            dest="inActivityTimeout",
            type=int,
            default=0)

    parser.add_argument(
            "--hard-timeout",
            help="Automatically stops logging after n seconds (in seconds, default: 0 = disabled)",
            dest="hardTimeout",
            type=int,
            default=0)


    parser.add_argument(
            "--values-to-log",
            help="List of value-names to log (default: [" + ", ".join(DEFAULT_VALUES_TO_LOG) + "])",
            dest="valuesToLog",
            action='append',
            type=str,
            default=DEFAULT_VALUES_TO_LOG)

    parser.add_argument(
            "-f",
            "--fill-sample-gaps",
            help="Log all (active) flows even if no new sample was recorded - forces rewrites of previous values (default: false)",
            action="store_true",
            dest="fillSampleGaps",
            default=False)

    guiOptGroup = parser.add_argument_group('gui')
    guiOptGroup.add_argument(
            "--refresh-rate-live-stats",
            help="Live stats refresh rate (in seconds, default: " + str(DEFAULT_REFRESH_RATE_LIVE_STATS) + ")",
            dest="refreshRateLiveStats",
            type=float,
            default=DEFAULT_REFRESH_RATE_LIVE_STATS)

    guiOptGroup.add_argument(
            "--show-ephemeral-flows",
            help="Show ephemeral/short-living flows (default: false)",
            dest="showEphemeralFlows",
            action="store_true",
            default=False)

    # guiOptGroup.add_argument(
    #         "--ip2host",
    #         help="Tries to lookup hostname by IP address (default: false)",
    #         dest="translateIp2Hostname",
    #         action="store_true",
    #         default=False)

    guiOptGroup.add_argument(
            "--gui-help",
            help="Show keybinds and information about the GUI.",
            dest="showGuiHelp",
            action="store_true",
            default=False)

    # Filter
    filterOptGroup = parser.add_argument_group('filter')
    filterOptGroup.add_argument(
            "--filter-dst-port",
            help="Filter by destination port. Multiple occurrences possible",
            dest="filterDstPorts",
            action='append',
            type=int,
            default=[])

    filterOptGroup.add_argument(
            "--filter-src-port",
            help="Filter by source port. Multiple occurrences possible",
            dest="filterSrcPorts",
            action='append',
            type=int,
            default=[])

    filterOptGroup.add_argument(
            "--whitelisting",
            help="Enable whitelisting (default: false = blacklisting)",
            dest="filterWhiteListing",
            action='store_true',
            default=False)

    # misc options
    parser.add_argument(
            "-q",
            "--quiet", help="Whether to ouput anything at all (default: false)",
            action="store_true",
            dest="quiet",
            default=False)

    parser.add_argument(
            "--debug",
            help="Debug mode (default: false)",
            action="store_true",
            dest="debug",
            default=False)

    parser.add_argument(
            "--version",
            help="Print version information",
            action="store_true",
            dest="showVersion",
            default=False)

    options = parser.parse_args()

    if(options.outputBackend == "stdout"):
        options.guiBackend = None
        # highest priority --> in case of StdoutOutput no gui at all
    elif(options.inputBackend == "file"):
        # if reading from regular file --> use progressGui
        options.guiBackend = "progress"
    else:
        if(options.quiet):
            options.guiBackend = None
        else:
            options.guiBackend = "curses"

    return options

def startupSelfCheck(options):
    # test for root-permissions
    if (os.getuid() == 0):
        Utility.eprint("Error: Please do not run this program with root-privileges. Exiting...")
        sys.exit(1)


def printGuiHelp():
    print("\
Available keybinds:\n\
    'q':            quit TCPlog \n\
    '+':            increase time between gui refreshes \n\
    '-':            decrease time between gui refreshes \n\
    'arrow_right':  scroll flow table to the right\n\
    'arrow_left':   scroll flow table to the left\n\
    'arrow_top':    scroll flow table upwards\n\
    'arrow_down':   scroll flow table downwards\n\
    'page_up':      scroll flow table upwards (fast)\n\
    'page_down':    scroll flow table downwards (fast)\n\
"
)
    # TODO: print info about GUI-elements (abbrevations and so on...)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from .output_base import OutputBase

class StdoutOutput(OutputBase):
    def __init__(self, options, infoRegistry):
        self.options = options
        self.infoRegistry = infoRegistry
        self.__loggedSampleCounter = 0

    def setSampleQueue(self, sampleQueue):
        self.sampleQueue = sampleQueue

    def setValuesToLog(self, valuesToLog):
        self.valuesToLog = valuesToLog

    def tearDown(self):
        pass

    def startUp(self):
        pass

    def startupCheck(self):
        pass

    def preLogging(self):
        logLine = "###" + ' '.join(self.valuesToLog)
        print(logLine)

    def logSample(self):
        """Print log data to stdout"""
        try:
            sampleObject = self.sampleQueue.popleft()
        except IndexError:
            pass
        else:
            valueList = []
            for val in self.valuesToLog:
                valueList.append(str(getattr(sampleObject, val, "_")))
            logLine = ' '.join(valueList)
            print(logLine)
            self.__loggedSampleCounter = self.__loggedSampleCounter + 1

    def postLogging(self):
        print("### samples logged: " + str(self.__loggedSampleCounter))



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
from ..backend_base import BackendBase

class OutputBase(BackendBase, metaclass=ABCMeta):
    """
    Base template for all output backends.
    """

    @abstractmethod
    def preLogging(self):
        """
        Executed before any samples are logged (usage example: write meta-data about log)
        """
        pass

    @abstractmethod
    def logSample(self):
        """
        Logs a single sample from the sample queue.
        """
        pass

    @abstractmethod
    def postLogging(self):
        """
        Executed after all samples are logged (usage example: write summary about logging-session)
        """
        pass



# -*- coding: utf-8 -*-

import socket
import threading
import time
from collections import deque

SERVER_ADDRESS = "0.0.0.0"

class BroadcastLogServer(object):
    """Broadcasts log data to each client connected via socket"""

    def __init__(self, options, logSamples, infoRegistry):
        """Initializes data structures"""
        self.options = options
        self.infoRegistry = infoRegistry
        self.__stopped = threading.Event()

        self.__logSamples = logSamples
        self.__clients = {} #queue per client
        self.__host = SERVER_ADDRESS
        self.__port = self.options.logPort

        self.__updateClientQueuesThread = threading.Thread(target=self.updateClientQueues)
        self.__updateClientQueuesThread.daemon = True
        self.__updateClientQueuesThread.start()

        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 512)
        self.__sock.bind((self.__host, self.__port))

    def initTearDown(self):
        self.__stopped.set()

    def appendNewSample(self, newSample):
        self.__logSamples.append(newSample)

    def updateClientQueues(self):
        """Update output-queue for each socket-client"""
        while(not self.__stopped.wait(0.0001)):
            try:
                sampleObject = self.__logSamples.popleft()
            except IndexError:
                time.sleep(0.0001)
            else:
                for clientID in self.__clients:
                    self.__clients[clientID].append(sampleObject)


    def getNumberOfClients(self):
        """Returns number of clients currently connected"""
        return len(self.__clients)

    def listen(self):
        """Listens for new clients. Starts thread to send data for each new client"""
        self.__sock.listen(50)
        while(not self.__stopped.wait(0.0001)):
            try:
                client, address = self.__sock.accept()
            except:
                pass
            else:
                ip, port = address
                clientID = str(ip) + ":" + str(port)
                self.__clients[clientID] = deque()
                threading.Thread(target = self.sendToClient, args = (client,address, clientID), daemon=True).start()


    def sendToClient(self, client, address, clientID):
        """Performs the actual send operation of log data"""

        if(self.options.debug):
            print("New client connected: " + str(clientID))

        while(not self.__stopped.wait(0.0001)):
            try:
                sampleObject = self.__clients[clientID].popleft()
            except IndexError:
                time.sleep(0.0001)
            else:
                try:
                    valueList = []
                    for val in self.options.valuesToLog:
                        valueList.append(str(getattr(sampleObject, val, None)))
                    logLine = ' '.join(valueList)
                    client.sendall(bytes(logLine + "\n", 'UTF-8'))
                    time.sleep(0.0001)
                except:
                    if(self.options.debug):
                        print("Client disconnected or timedout: " + str(clientID))
                    del self.__clients[clientID]
                    client.close()
                    return False
        client.close()



# -*- coding: utf-8 -*-

import sys
import time

from ...utils.utilty import Utility
from .input_base import InputBase
from ...sample import Sample

class TcpInfoInput(InputBase):
    def __init__(self, options, infoRegistry):
        self.options = options
        self.infoRegistry = infoRegistry
        self.__loggedSampleCounter = 0
        self.startTimestamp = time.time()
        self.shutdown = False

    def startupCheck(self):
        try:
           import tcpinfo
           tcpinfo.getListOfAvailableValues()
        except:
            Utility.eprint(\
                "Error: tcpinfo module not found. Exiting..."\
                "\n\n"\
                "Please install the tcpinfo module\n"
                "OR use \"tcpprobe\" as input-backend instead"\
            )
            sys.exit(1)
        else:
            self.tcpinfo = tcpinfo


    def startUp(self):
        self.tcpinfo.startUp()
        self.availableValues = self.tcpinfo.getListOfAvailableValues()

    def tearDown(self):
        self.shutdown = True
        self.tcpinfo.tearDown()
        pass

    def retrieveNewSamples(self):
        newSamples = []

        if(self.shutdown):
            return newSamples
        else:
            for flowSample in self.tcpinfo.getTcpInfoList():
                try:
                    if(flowSample['state'] != 'ESTABLISHED'): # only parse ESTABLISHED connections
                        continue
                except KeyError:
                        continue
                else:
                    dstPort = int(flowSample['dstPort'])
                    srcPort = int(flowSample['srcPort'])
                    log = False

                    if(self.options.filterWhiteListing):
                        log = False
                        if((srcPort in self.options.filterSrcPorts) or (dstPort in self.options.filterDstPorts)):
                            log = True

                    else: # blacklisting
                        log = True
                        if(srcPort in self.options.filterSrcPorts or dstPort in self.options.filterDstPorts):
                            log = False



                    if(log):
                        relativeTimestamp = time.time() - self.startTimestamp
                        sample = Sample(flowSample['srcIp'], flowSample['srcPort'], flowSample['dstIp'], flowSample['dstPort'])
                        sample.relativeTimestamp = relativeTimestamp

                        for value in self.availableValues:
                            if(value == "bytes_acked"):
                                sample.seq = flowSample['bytes_acked']
                            elif(value == "snd_ssthresh"):
                                sample.sst = flowSample['snd_ssthresh']
                            elif(value == "rtt"):
                                sample.rtt = float(flowSample['rtt']) / 1000
                            elif(value == "rcv_rtt"):
                                sample.rcv_rtt = float(flowSample['rcv_rtt']) / 1000
                            elif(value == "rttvar"):
                                sample.rttvar = float(flowSample['rttvar']) / 1000
                            else:
                                try:
                                    setattr(sample, value, flowSample[value])
                                except KeyError:
                                    pass

                        newSamples.append(sample)

                    else:
                        # skip sample
                        pass

            return newSamples



# -*- coding: utf-8 -*-

import threading
from collections import deque

from .output_base import OutputBase
from .socket_broadcast_log_server import BroadcastLogServer

THREAD_JOIN_TIMEOUT = 1 # in seconds

class SocketOutput(OutputBase):
    def __init__(self, options, infoRegistry):
        self.options = options
        self.infoRegistry = infoRegistry
        self.tmpSampleQueue = deque()
        self.__socketServer = None

    def setSampleQueue(self, sampleQueue):
        self.sampleQueue = sampleQueue

    def setValuesToLog(self, valuesToLog):
        self.valuesToLog = valuesToLog

    def startUp(self):
        pass

    def startupCheck(self):
        pass

    def tearDown(self):
        if self.__socketServer is not None:
            self.__socketServer.initTearDown()
        self.__initSocketServerThread.join(THREAD_JOIN_TIMEOUT)

    def preLogging(self):
        self.__initSocketServerThread = threading.Thread(target=self.initSocketServer)
        self.__initSocketServerThread.daemon = True
        self.__initSocketServerThread.start()

    def initSocketServer(self):
        self.__socketServer = BroadcastLogServer(self.options, self.tmpSampleQueue, self.infoRegistry).listen()

    def logSample(self):
        try:
            sampleObject = self.sampleQueue.popleft()
        except IndexError:
            pass
        else:
            self.tmpSampleQueue.append(sampleObject)

    def postLogging(self):
        self.tearDown()





# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
from ..backend_base import BackendBase

class InputBase(BackendBase, metaclass=ABCMeta):
    """
    Base template for all input backends.
    """

    @abstractmethod
    def retrieveNewSamples(self):
        """
        Returns list of new samples.
        """
        pass



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod

class BackendBase(metaclass=ABCMeta):
    """
    Base template for all backends
    """

    @abstractmethod
    def startupCheck():
        """
        Every backend must be able to determine if all of their necessary dependencies are met
        """
        pass

    @abstractmethod
    def startUp():
        """
        Initialize backend's dependencies (eg: file opening...)
        """
        pass

    @abstractmethod
    def tearDown():
        """
        Executed at programme shutdown: Every backend must perform a cleanup (eg: close files...)
        """
        pass



# -*- coding: utf-8 -*-

import time

DERIVED_VALUES = ['minRtt', 'maxRtt', 'avgRtt', 'meanRtt', 'throughput', 'smoothedThroughput', 'assumedLosses', 'transferedData']
DERIVED_VALUES_UNITS = ['ms', 'ms', 'ms', 'ms', 'byte/s', 'byte/s', '#', 'byte']
ACTIVITY_THRESHOLD = 10 # seconds
SMOOTHED_THROUGHPUT_THRESHOLD = 1 #second
EPHEMERAL_THRESHOLD = 2 # number of samples

class Flow(object):
    """Data structure to represent a filtered flow"""

    def __init__(self, flowIdentifier, srcIp, srcPort, dstIp, dstPort, sample):
        """Initializes new flow"""
        self.flowIdentifier = flowIdentifier
        self.srcIp = srcIp
        self.srcPort = srcPort
        self.dstIp = dstIp
        self.dstPort = dstPort
        self.srcHostname = None
        self.dstHostname = None
        # self.flowDirection = "in"
        # self.flowDirection = "out"
        self.timestampOfFirstActivity = time.time()
        self.timestampOfLastActivity = self.timestampOfFirstActivity
        self.currentSample = sample
        self.previousSample = sample
        self.timeBetweenCurrentAndPreviousSample = 0
        self.minRtt = sample.rtt
        self.maxRtt = sample.rtt
        self.avgRtt = sample.rtt
        self.meanRtt = sample.rtt
        self.rttSum = sample.rtt
        self.transferedData = 0
        self.throughput = 0
        self.smoothedThroughput = 0
        self.assumedLosses = 0
        self.numberOfSamples = 1
        self.logStatus = False
        self.timestampOfLastThroughputCalc = sample.relativeTimestamp
        self.seqLastThroughputCalc = sample.seq
        self.seqOfFirstActivity = sample.seq

    def updateFlow(self, newSample):
        self.logStatus = False
        self.numberOfSamples += 1
        self.rttSum += newSample.rtt
        self.timestampOfLastActivity = time.time()
        self.previousSample = self.currentSample
        self.currentSample = newSample
        self.timeBetweenCurrentAndPreviousSample = self.currentSample.absoluteTimestamp - self.previousSample.absoluteTimestamp
        self.updateDerivedValues()
        self.appendDerivedValuesToCurrentSample()

    def updateDerivedValues(self):
        # calculate misc rtt-vals
        self.minRtt = min(self.minRtt, self.currentSample.rtt)
        self.maxRtt = max(self.maxRtt, self.currentSample.rtt)
        self.avgRtt = self.rttSum / self.numberOfSamples
        #TODO: rename meanRtt to smoothedRtt?
        self.meanRtt = 0.01 * self.currentSample.rtt + 0.99 * self.meanRtt

        # NOTE: Packet loss is assumed, when SSThresh is decreased (significantly)
        # --> It was observed that TCP-Vegas repeatedly reduces SSThresh by a small amount, e.g. after Slow-Start (under some circumstances)
        if(self.previousSample.sst > (self.currentSample.sst + min(int(self.previousSample.sst) * 0.2, 10)) ):
            self.assumedLosses += 1

        # calculate throughput
        self.throughput = (self.currentSample.seq - self.previousSample.seq) if (self.currentSample.seq - self.previousSample.seq) >= 0 \
            else (int("0xffffffff",0) - self.previousSample.seq + self.currentSample.seq)
        self.throughput /= self.timeBetweenCurrentAndPreviousSample  # byte/s
        self.throughput *= 8 / 1000 / 1000 # mbit/s

        # calculate smoothed throughput (to minify burst fragments)
        tpCalcDelta = self.currentSample.relativeTimestamp - self.timestampOfLastThroughputCalc
        if(tpCalcDelta >= SMOOTHED_THROUGHPUT_THRESHOLD):
            throughput = (self.currentSample.seq - self.seqLastThroughputCalc) if (self.currentSample.seq - self.seqLastThroughputCalc) >= 0 \
                else (int("0xffffffff",0) - self.seqLastThroughputCalc + self.currentSample.seq)
            throughput /= tpCalcDelta  # byte/s
            throughput *= 8 / 1000 / 1000 # mbit/s
            self.smoothedThroughput = throughput
            self.timestampOfLastThroughputCalc = self.currentSample.relativeTimestamp
            self.seqLastThroughputCalc = self.currentSample.seq

        self.transferedData = self.currentSample.seq - self.seqOfFirstActivity
        # TODO:
        # IN- vs OUT-data
        # self.transferedOutData = self.currentSample.seq - self.seqOfFirstActivity
        # self.transferedInData = self.currentSample.bytesAcked - self.seqOfFirstActivity


    def appendDerivedValuesToCurrentSample(self):
        for val in DERIVED_VALUES:
            valueContent = (getattr(self, val, "_"))
            setattr(self.currentSample, val, valueContent)

    def timeSinceLastActivity(self):
        return (time.time() - self.timestampOfLastActivity)

    def totalActiveTime(self):
        return (self.timestampOfLastActivity - self.timestampOfFirstActivity)

    def getListOfDerivedValues(self):
        return DERIVED_VALUES

    def getListOfDerivedValuesUnits(self):
        return DERIVED_VALUES_UNITS

    def isActive(self):
        return (self.timeSinceLastActivity() < ACTIVITY_THRESHOLD)

    def isEphemeral(self):
        """
        If there is only a single sample (yet) count flow as short-lived.
        """
        return (self.numberOfSamples < EPHEMERAL_THRESHOLD)

    def getLogStatus(self):
        return self.logStatus

    def setLogStatus(self, logStatus):
        self.logStatus = logStatus

    def isIngressFlow(self):
        return (self.flowDirection == "in")

    def isEgressFlow(self):
        return (self.flowDirection == "out")



# -*- coding: utf-8 -*-

import time

class Sample(object):
    """Data structure ro represent a single log sample object"""

    def __init__(self, srcIp, srcPort, dstIp, dstPort):
        """Initializes a new sample object"""

        self.absoluteTimestamp = time.time()
        self.srcIp = srcIp
        self.srcPort = srcPort
        self.dstIp = dstIp
        self.dstPort = dstPort
        self.src = str(self.srcIp) + ":" + str(self.srcPort)
        self.dst = str(self.dstIp) + ":" + str(self.dstPort)
        self.seq = 0
        self.sst = 0

    def getFlowIdentifier(self):
        return self.src + "-" + self.dst




# -*- coding: utf-8 -*-

# TODO move logic from tcplog.py to here

class Flows(object):
    """
    Holds all flow objects and provides meta-stats about them.
    """

    def __init__(self):
        self.flows = {}

    def addFlow(self, flowIdentifier, flow):
        self.flows[flowIdentifier] = flow

    def getFlow(self, flowIdentifier):
        return self.flows[flowIdentifier]

    def getFlows(self):
        return self.flows

    def numberOfActiveFlows(self):
        numberOfActiveFlows = 0
        for flowIdentifier in self.flows:
            if self.flows[flowIdentifier].isActive():
                numberOfActiveFlows += 1

        return numberOfActiveFlows

    def getActiveFlows(self):
        # TODO
        pass

    def getLongLivingFlows(self):
        # TODO
        # {k: v for k, v in points.items() if v[0] < 5 and v[1] < 5}
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup

# keep in sync with main.py (!)
version = "0.2.7"
desc = "Tool to log and analyze TCP flows."

setup(
        name = "tcplog",
        packages = [
            "tcplog",
            "tcplog.utils",
            "tcplog.backends",
            "tcplog.backends.input",
            "tcplog.backends.output",
            "tcplog.backends.gui"
            ],
        entry_points = {
            "console_scripts": [
                'tcplog = tcplog.main:main'
                ],
            },
        version = version,
        description = desc,
        long_description = desc,
        author = "Karlsruhe Institute of Technology - Institute of Telematics",
        author_email = "telematics@tm.kit.edu",
        maintainer = "Michael Koenig",
        maintainer_email = "michael.koenig2@student.kit.edu",
        url = "https://git.scc.kit.edu/CPUnetLOG/TCPlog/",
        license = "BSD",
        platforms = "Linux",
        zip_safe = False,
        install_requires = [
            'tcpinfo>=0.2.1'#,
            # 'netifaces>=0.10.4'

        ],
        extras_require = {
            'Live visualization':  ["tcpliveplot>=0.2"]
        },
        keywords = ['tcp', 'flow', 'log', 'analyze', 'network', 'traffic'],
        classifiers = [
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Operating System :: POSIX :: Linux',
            'Environment :: Console',
            'Environment :: Console :: Curses',
            'Natural Language :: English',
            'Intended Audience :: Education',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',
            'Intended Audience :: System Administrators',
            'Intended Audience :: Telecommunications Industry',
            'Topic :: Scientific/Engineering',
            'Topic :: Internet',
            'Topic :: System :: Logging',
            'Topic :: System :: Networking',
            'Topic :: System :: Networking :: Monitoring',
            'Topic :: System :: Operating System Kernels :: Linux',
            'Topic :: Utilities'
        ]
        )



# -*- coding: utf-8 -*-

import sys
import os
import datetime

from ...utils.utilty import Utility
from .output_base import OutputBase

TCP_LOG_FORMAT_VERSION = "2"

# TODO: get global info from tcplog

class FileOutput(OutputBase):
    def __init__(self, options, infoRegistry):
        self.options = options
        self.infoRegistry = infoRegistry
        self.startTime = self.infoRegistry.load('startTime')
        self.__loggedSampleCounter = 0
        self.__logFileHandler = None

    def setSampleQueue(self, sampleQueue):
        self.sampleQueue = sampleQueue

    def setValuesToLog(self, valuesToLog):
        self.valuesToLog = valuesToLog

    def startUp(self):
        try:
            self.__logFileHandler = open(self.options.logFilePath, 'w')
        except:
            Utility.eprint("Error: while opening log file: " + self.options.logFilePath)
            sys.exit(1)

    def startupCheck(self):
        pass

    def tearDown(self):
        self.__logFileHandler.close()
        pass

    def preLogging(self):
        # open file and clear content (if there is any)
        os.chmod(self.options.logFilePath, 0o644)
        self.__logFileHandler.truncate(0)

        startTime = str(self.startTime)
        timeAsString = datetime.datetime.fromtimestamp(self.startTime).strftime('%Y-%m-%d %H:%M:%S')

        # write connections-stats
        self.__logFileHandler.write("#TCPLOG\n")
        self.__logFileHandler.write("#VERSION: " + str(self.infoRegistry.load("logFormatVersion")) + "\n")
        self.__logFileHandler.write("#DATE: " + timeAsString + "\n")
        self.__logFileHandler.write("#TIMESTAMP: " + startTime + "\n")
        self.__logFileHandler.write("#RESOLUTION: " + str(self.options.logResolution) + "s" + "\n")
        self.__logFileHandler.write("#FORMAT: " + " ".join(map(str, self.options.valuesToLog)) +"\n")
        # self.__logFileHandler.write("#UNITS: " + " ".join(map(str, VALUES_TO_LOG_UNITS)) +"\n")
        self.__logFileHandler.write("#END_HEADER\n\n")



        # logLine = "###" + ' '.join(self.valuesToLog)
        #print(logLine)
        pass

    def logSample(self):
        try:
            sampleObject = self.sampleQueue.popleft()
        except IndexError:
            pass
        else:
            valueList = []
            for val in self.valuesToLog:
                valueList.append(str(getattr(sampleObject, val, "_")))
            logLine = ' '.join(valueList)
            self.__logFileHandler.write(logLine + "\n")
            self.__loggedSampleCounter = self.__loggedSampleCounter + 1
        pass

    def postLogging(self):
        self.__logFileHandler.close()
        #print("### samples logged: " + str(self.__loggedSampleCounter))
        pass
